//----------------------------------------------------------------------------
//
// Dlq - a Darklands Quest Lister
// (C) 2000 by Michael Petzold
//
//----------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <dir.h>
#include <assert.h>

#include "DlqDump.h"


#define QUESTSZ        48
#define LOCSZ          58
#define LOCNAMESZ      20
#define CHARSZ        554
#define ITEMSZ         46
#define BASIC_OFFSET  393

#define QUESTNUM_OFFSET     (BASIC_OFFSET + (num_chars * CHARSZ))
#define FIRST_QUEST_OFFSET  (QUESTNUM_OFFSET + 2)
#define QUESTX_OFFSET(x)    (FIRST_QUEST_OFFSET + ((x) * QUESTSZ))
#define LOCNUM_OFFSET       (FIRST_QUEST_OFFSET + (num_quests * QUESTSZ))

#define MAX_LOC 420

char heading[512];

static FILE *game, *ifile;
static char loc_buf[MAX_LOC][LOCSZ];
static short num_chars, num_quests;

enum { N, NE, E, SE, S, SW, W, NW };

/****************************************************************************/

int find_most_recent_save (char *fnum)
{
    unsigned dte, tme;
    struct ffblk fb;

    fnum[2] = '\0';

    if (findfirst (".\\SAVES\\DKSAVE*.SAV", &fb, 0) == 0) {
        dte = fb.ff_fdate;
        tme = fb.ff_ftime;
        fnum[0] = fb.ff_name[6];
        if (fb.ff_name[7] != '.') {
            fnum[1] = fb.ff_name[7];
        }
        else {
            fnum[1] = '\0';
        }
    }
    else {
      sprintf (error_text, "Couldn't find a savegame (.\\SAVES\\DKSAVE0.SAV).");
      return 0;
    }

    while (findnext(&fb) == 0) {
        if ((fb.ff_fdate > dte) ||
            ((fb.ff_fdate == dte) && (fb.ff_ftime > tme))) {
            dte = fb.ff_fdate;
            tme = fb.ff_ftime;
            fnum[0] = fb.ff_name[6];
            if (fb.ff_name[7] != '.') {
                fnum[1] = fb.ff_name[7];
            }
            else {
                fnum[1] = '\0';
            }
        }
    }

  return 1;
}

/****************************************************************************/

static void dl_to_win (char *name)
{
    /* DL's town names only have '' and '' umlauts */
    while (*name) {
        if (*name == 0x7b) *name = 0xf6;          /* : DOS-0x94 ANSI-0xf6 */
        else if (*name == 0x7c) *name = 0xfc;     /* : DOS-0x81 ANSI-0xfc */
        name++;
    }
}

/****************************************************************************/

int get_file (const char *fname, const char *fnum)
{
    short num_loc, year, month, day, hour;
    unsigned char nc;
    char t1[80], t2[80];

    if ((game = fopen (fname, "rb")) == NULL) {
        sprintf (error_text, "Couldn't open \"");
        sprintf (error_text + strlen (error_text), fname);
        sprintf (error_text + strlen (error_text), "\".");
        return 0;
    }
    if ((ifile = fopen ("darkland.lst", "rb")) == NULL) {
        sprintf (error_text, "Couldn't open \"darkland.lst\".");
        fclose (game);
        return 0;
    }

    fseek (game, 104, SEEK_SET);
    fread (&year, 2, 1, game);
    fread (&month, 2, 1, game);
    fread (&day, 2, 1, game);
    fread (&hour, 2, 1, game);
    month++;
    fseek (game, 241, SEEK_SET);
    fread (&nc, 1, 1, game);
    num_chars = (short) nc;

    fseek (game, QUESTNUM_OFFSET, SEEK_SET);
    fread (&num_quests, 2, 1, game);
    fseek (game, LOCNUM_OFFSET, SEEK_SET);
    fread (&num_loc, 2, 1, game);
    if (num_loc > MAX_LOC) {
        num_loc = MAX_LOC;
    }
    fread (&loc_buf[0], LOCSZ, num_loc, game);

    rewind (game);
    fread (t1, 21, 1, game);
    dl_to_win (t1);
    fread (t2, 80, 1, game);
    dl_to_win (t2);
    sprintf (heading, "Quests for DKSAVE%s.SAV: %s %s %hi/%hi/%hi hour %hi",
            fnum, t1, t2, year, month, day, hour);

    return 1;
}

/****************************************************************************/

int create_fname_array (MyStringArray *namearray)
{
    struct ffblk fb;
    int rslt;
    FILE *thisgame;
    short year, month, day, hour;
    char fname[81], fnum[3], t1[80], t2[80], fileline[512];

    if ((rslt = findfirst (".\\SAVES\\DKSAVE*.SAV", &fb, 0)) != 0) {
        sprintf (error_text, "Couldn't find a savegame (.\\SAVES\\DKSAVE0.SAV).");
        return 0;
    }

    while (rslt == 0) {
        strcpy(fname, ".\\SAVES\\");
        strcat (fname, fb.ff_name);
        fnum[2] = '\0';
        fnum[0] = fb.ff_name[6];
        if (fb.ff_name[7] != '.')
            fnum[1] = fb.ff_name[7];
        else
            fnum[1] = '\0';
        if ((thisgame = fopen (fname, "rb")) == NULL) {
            sprintf (error_text, "Couldn't open \"");
            sprintf (error_text + strlen (error_text), fname);
            sprintf (error_text + strlen (error_text), "\".");
            return 0;
        }
        fread (t1, 21, 1, thisgame);
        fread (t2, 80, 1, thisgame);
        fseek (thisgame, 104, SEEK_SET);
        fread (&year, 2, 1, thisgame);
        fread (&month, 2, 1, thisgame);
        fread (&day, 2, 1, thisgame);
        fread (&hour, 2, 1, thisgame);
        dl_to_win (t1);
        dl_to_win (t2);
        month++;                            /* DL peculiarity... */
        sprintf (fileline, " DKSAVE%s.SAV: %s %s %hi/%hi/%hi  hour %hi",
                 fnum, t1, t2, year, month, day, hour);
        namearray->Add(fileline);
        fclose (thisgame);
        rslt = findnext(&fb);
    }

  return 1;
}

/****************************************************************************/

static void get_item_desc (char *desc, short num)
{
    fseek (ifile, 3 + (num * ITEMSZ), SEEK_SET);
    fread (desc, ITEMSZ, 1, ifile);
}

/****************************************************************************/

static void get_location (char *name, short num)
{
    if (num >= MAX_LOC) {
        strcpy (name, "<invalid index>");
        return;
    }
    strncpy (name, &loc_buf[num][38], LOCNAMESZ);
    name[LOCNAMESZ] = '\0';
    dl_to_win (name);
}

/****************************************************************************/

static int get_dir (int site_x, int site_y, int city_x, int city_y)
{
    double angle, dx, dy;
    int dir;

    dx = (double) abs (site_x - city_x);
    dy = (double) abs (site_y - city_y);
    if (dx == 0.0)
        dx = 0.0001;                        /* atan2 uses dy/dx */
    angle = atan2 (dy, dx);

    if (angle < (M_PI / 8.0)) {             /* 0,3926990816987 (22.5) */
        if (site_x > city_x)
            dir = E;
        else
            dir = W;
    }
    else if (angle < (M_PI * 3.0 / 8.0)) {  /* 1,178097245096 (67.5) */
        if (site_x > city_x) {
            if (site_y < city_y)
                dir = NE;
            else
                dir = SE;
        }
        else {
            if (site_y > city_y)
                dir = SW;
            else
                dir = NW;
        }
    }
    else {
        if (site_y < city_y)
            dir = N;
        else
            dir = S;
    }

    return (dir);
}

/****************************************************************************/

static short find_closest_city (short num, int *dir, short closest)
{
    double x_diff, y_diff, dist_min, dist;
    short x, y, x2, y2, save_x, save_y, i, where;

    dist_min = 99999.0L;
    where = -1;

    x = ((short *) loc_buf[num])[2];
    y = ((short *) loc_buf[num])[3];

    for (i = 0; i < 92; i++) {
        if (i == closest)
            continue;
        x2 = ((short *) loc_buf[i])[2];
        y2 = ((short *) loc_buf[i])[3];
        x_diff = fabs ((double) x2 - (double) x);
        y_diff = fabs ((double) y2 - (double) y);
        dist = sqrt ((x_diff * x_diff) + (y_diff * y_diff));
        if (dist < dist_min) {
            dist_min = dist;
            save_x = x2;
            save_y = y2;
            where = i;
        }
    }

    *dir = get_dir (x, y, save_x, save_y);
    return (where);
}

/****************************************************************************/

int dumpit (MyStringArray *bodyarray)
{
    const char * const who[] = {"Merchant", "#1", "#2", "#3",
                                "Foreign Trader", "Pharmacist", "Medici",
                                "Hanseatic League", "Fugger", "Schulz", "Mayor",
                                "#11"};
    const char * const nsew[] = {"N", "NE", "E", "SE", "S", "SW", "W", "NW"};
    char item[ITEMSZ];
    char location[LOCNAMESZ + 1];
    char line1[512], line2[512];
    short quest[QUESTSZ / 2];
    int i, dd, mm, yy, d, m, y, director, pending_quests = 0;

    for (i = 0; i < num_quests; i++) {
        fseek (game, QUESTX_OFFSET (i), SEEK_SET);
        fread (&quest, QUESTSZ, 1, game);
        d = quest[6];
        m = quest[7];
        y = quest[8];
        dd = quest[10];
        mm = quest[11];
        yy = quest[12];

        if ((quest[13] < 0) || (yy == y)) { /* quest not pending? */
            /* this is a k.o. criterium */
            continue;
        }
        if ((quest[17] != 8) || (quest[21] == 0) || (quest[22] != 2)) {
            /* here we come only if it is_potentially invalid */
            if (quest[23] == 0) {
                /* here we are again sure that it is invalid -
                   valid 'return to xyz' entries arrive here, but for them
                   quest[23] != 0. */
                continue;
            }
        }

        sprintf (line1, "%3i: ", ++pending_quests);

        if (quest[23] != 0) {           /* item? */
            get_item_desc (item, quest[23]);
            if (yy == 1499)
                sprintf (line1 + strlen (line1), "Get %s from ", item);
            else
                sprintf (line1 + strlen (line1), "Return %s to %s at ", item,
                         who[quest[13]]);
        }
        else {
            sprintf (line1 + strlen (line1), "Raubritter Quest - go to - ");
        }

        get_location (location, quest[14]);      /* destination */
        sprintf (line1 + strlen (line1), "%s ", location);

        if (quest[14] > 91) {
            short where;
            where = find_closest_city (quest[14], &director, -1);
            get_location (location, where);
            sprintf (line1 + strlen (line1), "(%s of %s, ",
                     nsew[director], location);
            where = find_closest_city (quest[14], &director, where);
            get_location (location, where);
            sprintf (line1 + strlen (line1), "%s of %s) ",
                     nsew[director], location);
        }

        if (quest[12] > 1498)
            sprintf (line1 + strlen (line1),
                     "[from %4d/%02d/%02d]", y, m + 1, d);
        else
            sprintf (line1 + strlen (line1),
                     "by %4d/%02d/%02d ", yy, mm + 1, dd);

//---
#if 0
        if (quest[17] != 8) {
            sprintf (line1 + strlen (line1), " <[17] != 8>");
        }
        if (quest[21] == 0) {
            sprintf (line1 + strlen (line1), " <[21] == 0>");
        }
        if (quest[22] != 2) {
            sprintf (line1 + strlen (line1), " <[22] != 2>");
        }
#endif        
//---
        bodyarray->Add(line1);

        get_location (location, quest[15]);
        if (quest[23] != 0) {           /* item */
            if (yy == 1499) {           /* return it */
                sprintf (line2, "     Deliver to %s at %s",
                         who[quest[13]], location);
                bodyarray->Add(line2);
            }
        }
        else {                          /* rk quest */
            sprintf (line2, "     Report to %s at %s",
                     who[quest[13]], location);
            bodyarray->Add(line2);
        }

        bodyarray->Add(" ");
    }
    fclose (game);
    fclose (ifile);

    return pending_quests;
}

/****************************************************************************/

