//----------------------------------------------------------------------------
//
// Dlq - a Darklands Quest Lister
// (C) 2000-2001 by Michael Petzold
//
//----------------------------------------------------------------------------

#include <stdio.h>                          /* sprintf */
#include <string.h>                         /* strcat, strcpy */

#include <dos.h>                            /* getdfree */
#include <dir.h>                            /* findfirst, etc.*/

#include "DlqApp.h"                         /* error_text */
#include "DlqSort.h"


#define FILENAME_MAXLENGTH (13 + 1)

int num_files;
static char file_names[200][FILENAME_MAXLENGTH];
static long space_needed;
static unsigned short dte[200], tme[200];

/****************************************************************************/

static int load_file_name_list ()
{
    int i;
    struct ffblk fb;

    space_needed = 0L;
    i = 0;

    if (findfirst (".\\SAVES\\DKSAVE*.??V", &fb, 0) == 0) {
        strcpy (file_names[i], fb.ff_name);
        dte[i] = fb.ff_fdate;
        tme[i] = fb.ff_ftime;
        space_needed += fb.ff_fsize;
        ++i;
    }

    while (findnext (&fb) == 0 && i < 200) {
        strcpy (file_names[i], fb.ff_name);
        dte[i] = fb.ff_fdate;
        tme[i] = fb.ff_ftime;
        space_needed += fb.ff_fsize;
        ++i;
    }

    if (findnext (&fb) == 0) {
//        printf ("More than 200 files found (%i) - "
//                "only first 200 will be re-sorted\n", i);
    }

    return i;
}

/****************************************************************************/

static int enough_disk_space (long bytes_needed)
{
    struct dfree dt;
    double bytes_free;

    getdfree (0, &dt);                      /* 0: current drive */

    if (dt.df_sclus == (unsigned) -1) {     /* error condition */
        return -1;
    }

    bytes_free = ((double) dt.df_avail *    /* free clusters */
                  (double) dt.df_sclus *    /* sectors per cluster */
                  (double) dt.df_bsec);     /* bytes per sector */

/*    printf ("(%.0f bytes free on current drive)\n", bytes_free); */

    if (dt.df_avail <= num_files ||         /* at least one cluster per file */
        bytes_free <= bytes_needed) {
        return -1;
    }

    return 0;
}

/****************************************************************************/

static int copy_files_to_cwd ()
{
    bool r = TRUE;
    for (int i = 0; i < num_files; ++i) {
        char old_file[sizeof (".\\SAVES\\") + FILENAME_MAXLENGTH] = ".\\SAVES\\";
        char new_file[sizeof (".\\") + FILENAME_MAXLENGTH] = ".\\";
        strcat (old_file, file_names[i]);
        strcat (new_file, file_names[i]);
        r = CopyFile (old_file, new_file, TRUE);
        if (r != TRUE) {
            sprintf (error_text, "CopyFile failed.");
            break;
        }
    }
    return !r;
}

/****************************************************************************/

static void sort_files ()
{
    int i, j;
    unsigned short dd, tt;
    char tname[FILENAME_MAXLENGTH];                     /* DL uses DOS names */

    for (i = 0; i < (num_files + 1); i++) {
        for (j = (i + 1); j < num_files; j++) {
            if ((dte[j] > dte[i]) ||
                ((dte[j] == dte[i]) && (tme[j] > tme[i])) ||
                ((dte[j] == dte[i]) && (tme[j] == tme[i]) &&
                 (strcmp (file_names[j], file_names[i]) > 0 ))) {
                strcpy (tname, file_names[i]);
                strcpy (file_names[i], file_names[j]);
                strcpy (file_names[j], tname);
                dd = dte[i];
                dte[i] = dte[j];
                dte[j] = dd;
                tt = tme[i];
                tme[i] = tme[j];
                tme[j] = tt;
            }
        }
    }
}

/****************************************************************************/

static int delete_save_files ()
{
    bool r = TRUE;
    for (int i = 0; i < num_files; ++i) {
        char file[sizeof (".\\SAVES\\") + FILENAME_MAXLENGTH] = ".\\SAVES\\";
        strcat (file, file_names[i]);
        r = DeleteFile (file);
        if (r != TRUE) {
            sprintf (error_text, "DeleteFile failed.");
            break;
        }
    }
    return !r;
}

/****************************************************************************/

static int copy_files_to_savedir ()
{
    bool r = TRUE;
    for (int i = 0; i < num_files; ++i) {
        char old_file[sizeof (".\\") + FILENAME_MAXLENGTH] = ".\\";
        char new_file[sizeof (".\\SAVES\\") + FILENAME_MAXLENGTH] = ".\\SAVES\\";
        strcat (old_file, file_names[i]);
        strcat (new_file, file_names[i]);
        r = CopyFile (old_file, new_file, TRUE);
        if (r != TRUE) {
            sprintf (error_text, "CopyFile failed.");
            break;
        }
    }
    return !r;
}

/****************************************************************************/

static int delete_temp_files ()
{
    bool r = TRUE;
    for (int i = 0; i < num_files; ++i) {
        char file[sizeof (".\\") + FILENAME_MAXLENGTH] = ".\\";
        strcat (file, file_names[i]);
        r = DeleteFile (file);
        if (r != TRUE) {
            sprintf (error_text, "DeleteFile failed.");
            break;
        }
    }
    return !r;
}

/****************************************************************************/

int dlq_sort ()
{
    num_files = load_file_name_list ();

    if (num_files == 0) {
        sprintf (error_text, "No save games found in directory .\\SAVES\\\n");
        return 0;
    }
    if (enough_disk_space (space_needed + 2048) != 0) {
        sprintf (error_text, "Not enough disk space to sort.\n"
                 "(%ld bytes are needed)", space_needed + 2048);
        return 0;
    }
    if (copy_files_to_cwd () != 0)
        return 0;

    sort_files ();

    if (delete_save_files () != 0)
        return 0;
    if (copy_files_to_savedir () != 0)
        return 0;
    if (delete_temp_files () != 0)
        return 0;

    return 1;
}

/****************************************************************************/
