//----------------------------------------------------------------------------
//
// Dlq - a Darklands Quest Lister
// (C) 2000-2001 by Michael Petzold
//
//----------------------------------------------------------------------------

// Borland ObjectWindows 5.0 Library
#include <owl/scroller.h>

// C++ Run Time Library
#include <dir.h>                            /* chdir */

// Local includes
#include "DlqDump.h"
#include "DlqSort.h"
#include "FileDlg.h"
#include "Dlq.rc"
#include "DrawWindow.h"


static const short ShortKeyMask = 0x8000;


DEFINE_RESPONSE_TABLE1(DrawWindow, TWindow)
  EV_WM_SIZE,
  EV_WM_KEYDOWN,
  EV_COMMAND(CM_FILERELOAD, CmFileReload),
  EV_COMMAND(CM_FILEOPEN, CmFileOpen),
  EV_COMMAND(CM_FILESAVEAS, CmFileSaveAs),
  EV_COMMAND(CM_FILESORT, CmFileSort),
  EV_COMMAND(CM_ABOUT, CmAbout),
END_RESPONSE_TABLE;


DrawWindow::DrawWindow(TWindow *parent, TTextGadget* tg)
{
  Init(parent, 0, 0);
  BodyArray = new MyStringArray(10, 0, 10);
  FileData  = new TOpenSaveDialog::TData(OFN_HIDEREADONLY|OFN_FILEMUSTEXIST,
                                        "Text Files (*.txt)|*.txt|", 0, "",
                                        "txt");
  StatusText = tg;
  DocSizeValid = false;

  // Find a font that will fit into our window
  pFont1 = new TFont(
    "Times New Roman",          // facename
    HeightInPixels(9),          // height,
    0, 0, 0, FW_NORMAL,         // width, esc, orientation, weight
    VARIABLE_PITCH | FF_ROMAN,  // Pitch and Family
    false, false, false,        // Italic, Underline, Strikeout
    ANSI_CHARSET,               // Charset
    OUT_CHARACTER_PRECIS,       // Output precision
    CLIP_DEFAULT_PRECIS,        // Clip precision
    PROOF_QUALITY               // Quality
  );
  // This will paint the heading line
  pFont2 = new TFont(
    "Arial",                    // facename
    HeightInPixels(9),          // height,
    0, 0, 0, FW_BOLD,           // width, esc, orientation, weight
    VARIABLE_PITCH | FF_SWISS,  // Pitch and Family
    false, true, false,         // Italic, Underline, Strikeout
    ANSI_CHARSET,               // Charset
    OUT_CHARACTER_PRECIS,       // Output precision
    CLIP_DEFAULT_PRECIS,        // Clip precision
    PROOF_QUALITY               // Quality
  );

  rectUserMargins = TRect(20,20,40,40);
  GameDir = getcwd(NULL, MAXPATH);
}


DrawWindow::~DrawWindow()
{
  delete BodyArray;
  delete FileData;
  delete pFont1;
  delete pFont2;
  free(GameDir);
}


void DrawWindow::SetupWindow()
{
  TWindow::SetupWindow();               //Lets TWindow create any child controls

  // Scroll bars
  Attr.Style |= WS_VSCROLL | WS_HSCROLL;

  Scroller = new TScroller(this, 1, 1, 0, 0);
  AdjustScroller();

  while (!PostMessage(WM_COMMAND, CM_FILERELOAD));
}


void DrawWindow::ComputeScreenMetrics()
{
  string str;
  TScreenDC dc;
  TEXTMETRIC tm;
  MyStringIterator i(*BodyArray);
  TSize size;

  dc.SelectObject(*pFont1);

  // First metric: get size of logical inch in pixels
  sizeLogInch.cx = dc.GetDeviceCaps(LOGPIXELSX);
  sizeLogInch.cy = dc.GetDeviceCaps(LOGPIXELSY);

  dc.GetTextMetrics(tm);

  sizeCharScn.cy = tm.tmHeight + tm.tmExternalLeading;
  sizeCharScn.cx = tm.tmAveCharWidth;

  sizeSpaceScn.cy = sizeCharScn.cy / 2;
  sizeSpaceScn.cx = 0;

  sizeLineScn = TSize(0, 0);
  // loop through the document and find the longest line
  while (i) {
    str = i++;
    size = dc.GetTextExtent(str.c_str(), str.length());
    sizeLineScn.cx = max(size.cx, sizeLineScn.cx);
  }
  sizeLineScn.cy = sizeCharScn.cy;// + sizeSpaceScn.cy;

  dc.RestoreFont();
}


TSize DrawWindow::GetDocSize()
{
  ComputeScreenMetrics();

  TSize Size;
    Size.cx = sizeLineScn.cx + rectUserMargins.left + rectUserMargins.right;
    Size.cy = (BodyArray->GetItemsInContainer() * sizeLineScn.cy)
             + rectUserMargins.top + rectUserMargins.bottom;

  DocSize = Size;
  DocSizeValid = true;
  return Size;
}


void DrawWindow::AdjustScroller()
{
  TSize range = (DocSizeValid) ? DocSize : GetDocSize();

  // Use device unit for scroll range
  TClientDC dc(*this);
  dc.LPtoDP((TPoint*)&range);

  range -= GetClientRect().Size();
  Scroller->SetRange(range.cx, range.cy);
}


void DrawWindow::EvSize(UINT SizeType, TSize& Size)
{
  TWindow::EvSize(SizeType, Size);

  if (SizeType != SIZEICONIC) {
    AdjustScroller();
  }
}


void DrawWindow::EvKeyDown(uint key, uint repeatCount, uint flags)
{
  TSize range = (DocSizeValid) ? DocSize : GetDocSize();

  switch (key) {
      case
      VK_HOME : if (GetKeyState(VK_CONTROL) & ShortKeyMask)
                  Scroller->ScrollBy(-Scroller->XPos, -Scroller->YPos);
                else
                  Scroller->ScrollBy(0, -Scroller->YPos);
                break;
      case
      VK_END  : if (GetKeyState(VK_CONTROL) & ShortKeyMask)
                  Scroller->ScrollBy(range.cx - Scroller->XPos,
                                     range.cy - Scroller->YPos);
                else
                  Scroller->ScrollBy(0, range.cy - Scroller->YPos);
                break;
      case
      VK_PRIOR: Scroller->ScrollBy(0, -GetClientRect().Size().cy);
                break;
      case
      VK_NEXT:  Scroller->ScrollBy(0, GetClientRect().Size().cy);
                break;
      case
      VK_UP   : Scroller->ScrollBy(0, -10);
                break;
      case
      VK_DOWN : Scroller->ScrollBy(0, 10);
                break;
      case
      VK_LEFT : Scroller->ScrollBy(-10, 0);
                break;
      case
      VK_RIGHT: Scroller->ScrollBy(10, 0);
  }

  TWindow::EvKeyDown(key, repeatCount, flags);
}


void DrawWindow::Paint(TDC& dc, bool, TRect&)
{
  if (BodyArray->GetItemsInContainer() == 0)
    return;

  string s;
  int YPos = rectUserMargins.top;
  MyStringIterator i(*BodyArray);

  // heading
  dc.SelectObject(*pFont2);
  s = i++;
  dc.TextOut(rectUserMargins.left + 30, YPos, s.c_str(), s.length());
  YPos += (sizeCharScn.cy * 2);
  dc.RestoreFont();

  // quest list
  dc.SelectObject(*pFont1);
  if(!i) {
    s = "No quests so far...";
    dc.TextOut(rectUserMargins.left, YPos, s.c_str(), s.length());
  }
  else {
    while (i) {
      s = i++;
      dc.TextOut(rectUserMargins.left, YPos, s.c_str(), s.length());
      YPos += sizeLineScn.cy;
    }
  }
  dc.RestoreFont();
}


void DrawWindow::DumpFile(const char* fnum)
{
  char fname[81];

  strcpy(fname, ".\\SAVES\\DKSAVE");
  strcat(fname, fnum);
  strcat(fname, ".SAV");

  if(!get_file(fname, fnum)) {
    MessageBox(error_text, "File Error", MB_OK | MB_ICONEXCLAMATION);
    return;
  }

  BodyArray->Flush();
  Invalidate();
  if (StatusText)
    StatusText->SetText(NULL);

  char pname[90] = "Dlq - ";
  strcat(pname, heading + 11);
  Parent->SetCaption(pname);
  BodyArray->Add(heading);
  int nQuests = dumpit(BodyArray);
  if (StatusText) {
    char gadgettext[21];
    sprintf(gadgettext, "Total: %i Quests", nQuests);
    StatusText->SetText(gadgettext);
  }
  DocSizeValid = false;
  AdjustScroller();
}


void DrawWindow::CmFileReload()
{
  char fnum[3];

  if(!find_most_recent_save(fnum)) {
    MessageBox(error_text, "File Error", MB_OK | MB_ICONEXCLAMATION);
    return;
  }

  DumpFile(fnum);
}


void DrawWindow::CmFileOpen()
{
  char fnum[3];
  FileDlg dlg(this, fnum);

  if (dlg.Execute() != IDOK)
    return;

  if (fnum[0])
    DumpFile(fnum);
}


void DrawWindow::CmFileSaveAs()
{
  strcpy(FileData->FileName, "");

  if ((TFileSaveDialog(this, *FileData)).Execute() == IDOK)
    SaveFile();
  chdir(GameDir);
}


void DrawWindow::CmFileSort()
{
  if (MessageBox("Sorting save games now?\n\n",
                 "Sort save games",
                 MB_OKCANCEL | MB_ICONQUESTION) != IDOK)
    return;

  if (!dlq_sort())
    MessageBox(error_text, "File Error", MB_OK | MB_ICONEXCLAMATION);
  else
    MessageBox("Save games sorted.", "Success!", MB_OK | MB_ICONINFORMATION);
}


void DrawWindow::CmAbout()
{
  TDialog(this, IDD_ABOUT).Execute();
}


void DrawWindow::SaveFile()
{
  ofstream os(FileData->FileName);

  if (!os)
    MessageBox("Unable to open file", "File Error", MB_OK |
    MB_ICONEXCLAMATION);
  else {
    MyStringIterator i(*BodyArray);
    os << i++ << endl << endl;              // title
    while (i)
      os << i++ << endl;
  }
}



